/**
 * @file Ce fichier regroupe les différentes classes avec différents admins. Ces classes sont dédiées à être exportées directement pour être utilisées par le solver.
 * Le découpage par fichier est arbitraire mais permet de regrouper certaines classes proches.
 * @author hawkspar
 */

import {ldapConfig} from './config';
import {LDAP} from './basics';
import {Tests} from './utilities';
import { Open, User, userData, groupData } from './users';

/**
 * @interface userFullData
 * @desc Interface avec toutes les données d'un utilisateur. Etend {@link userData}.
 * @var {string} password - Mot de passe généré en amont
 * @arg {string} readPerm - Permissions spéciales BR
 * @var {string} writePerm - Permissions spéciales BR
 * @var {string[]} forlifes - Alias BR (attention le filtre .fkz n'est plus fonctionnel)
 * @var {string[]} groupsIsAdmin - Liste des gid dont le pax est admin ; supposé sous-liste de groups
 */
export interface userFullData extends userData {
    password: string,
    readPerm: string,
    writePerm: string,
    forlifes: string[],
    groupsIsAdmin: string[]
}

export class Admin extends User {
    /**
     * @class Cette classe est la classe de l'administrateur d'un groupe qui lui permet de rajouter des membres, en supprimer, idem pour des admins,
     * ou éditer, voir supprimer le groupe.
     * @summary Ce constructeur appelle simplement le constructeur de sa classe mère.
    */
    constructor() { super(); }
    
    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de relation TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui permet de rajouter un membre déjà créé à un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link LDAP.modifier} et {@link listerGroupes}. Elle n'autorise pas les doublons et opère dans les deux dns users
     * et groups.
     * @arg {string} uid - Identifiant du futur membre
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async addGroupMember(uid: string, gid: string) {
        try {
            // Vérifie que l'utilisateur est pas déjà membre pour groupes
            let lm = await Open.getMembers(gid);
            if (!lm.includes(uid)) {
                let vals = {};
                vals[ldapConfig.groups.member] = uid;
                // Erreur si pb lors de la modification
                if (!await LDAP.change(ldapConfig.key_id+gid+ldapConfig.dn_groups, "add", vals)) {
                    throw "Erreur lors de la modification dans l'arbre des groupes pour ajouter un membre.";
                }
            }
        }
        catch(err) {
            throw "Erreur lors de la recherche de la liste des membres pour ajouter un membre.";
        }
        try {
            // Vérifie que l'utilisateur est pas déjà membre pour users
            let lg = await Open.getGroups(uid);
            if (!lg.includes(gid)) {
                let vals2 = {};
                vals2[ldapConfig.users.groups] = gid;
                // Erreur si pb lors de la modification
                if (!await LDAP.change(ldapConfig.key_id+uid+ldapConfig.dn_users, "add", vals2)) {
                    throw "Erreur lors de la modification dans l'arbre des utilisateurs pour ajouter un membre.";
                }
            }
            return true;
        }
        catch(err) {
            throw "Erreur lors de la recherche de la liste des membres pour ajouter un membre.";
        }
    }

    /**
     * @summary Fonction qui permet de supprimer un membre existant d'un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link LDAP.search}, {@link LDAP.change}, {@link Open.getGroups} et {@link Open.getMembers}.
     * @arg {string} uid - Identifiant de l'ex-membre
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async delGroupMember(uid: string, gid: string) {
        try {
            // Vérifie que l'utilisateur est pas déjà viré pour groupes
            let lm = await Open.getMembers(gid);
            if (lm.includes(uid)) {
                // Supprime tous les utilisateurs
                if (!await LDAP.change(ldapConfig.key_id+gid+ldapConfig.dn_groups, "del", ldapConfig.group.member)) {
                    throw "Erreur lors de la suppression de tous les membres du groupe.";
                }
                // Les rajoute un par un, sauf pour le supprimé
                lm.forEach(id => {
                    if (id!=uid) {
                        this.addGroupMember(id, gid).then(res => {
                            if (!res) { throw "Erreur lors du ré-ajout des autres membres"; }
                        });
                    }
                });
            }
        }
        catch(err) {
            throw "Erreur pour obtenir une liste de membres d'un groupe pour supprimer un membre du groupe.";
        }
        try {
            let lg = await Open.getGroups(uid);
            // Vérifie que l'utilisateur est pas déjà viré pour users
            if (lg.includes(gid)) {
                // Supprime tous les groupes
                if (!await LDAP.change(ldapConfig.key_id+uid+ldapConfig.dn_users, "del", ldapConfig.member.groups)) {
                    throw "Erreur lors de la suppression de tous les groupes du membre.";
                }
                // Les rajoute un par un, sauf pour le supprimé
                lg.forEach(id => {
                    if (id!=gid) {
                        this.addGroupMember(uid, id).then(res => {
                            if (!res) { throw "Erreur lors du ré-ajout des autres groupes"; }
                        });
                    }
                });
            }
            return true;
        }
        catch(err) {
            throw "Erreur pour obtenir une liste de groupes d'un membres pour le supprimer du groupe.";
        }
    }

    /**
     * @summary Fonction qui permet de promouvoir membre au stade d'administrateur d'un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Admin.addGroupMember} {@link LDAP.change} et {@link Open.getAdmins}. Elle n'autorise pas
     * les doublons et opère dans les deux dns users et groups.
     * @arg {string} uid - Identifiant du futur membre
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async addGroupAdmin(uid: string, gid: string){
        // Ajoute le membre au groupe avant d'en faire un admin
        if (!await Admin.addGroupMember(uid,gid)) { throw "Erreur lors de l'ajout du futur admin en tant que membre."; }
        try {
            let la = await Open.getAdmins(gid);
            if (!la.includes(uid)) {
                // Finalement modification, uniquement dans groups
                let vals = {};
                vals[ldapConfig.groups.admin] = uid;
                if (!await LDAP.change(ldapConfig.key_id+gid+ldapConfig.dn_groups, "add", vals)) {
                    throw "Erreur lors de l'ajout de l'admin dans l'arbre des groupes.";
                }
            }
            return true;
        }
        catch(err) {
            throw "Erreur lors de l'obtention de la liste des administrateurs d'un groupe.";
        }
    }

    /**
     * @summary Fonction qui permet de rétrograder un membre du stade d'administrateur d'un groupe au stade d'utilisateur.
     * @desc Cette fonction fait essentiellement appel à {@link LDAP.change}.
     * @arg {string} uid - Identifiant du futur membre
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async delGroupAdmin(uid: string, gid: string) {
        // Peut paraître absurde mais permet de s'assurer que le membre est bien présent et que ses champs sont comme il faut
        if (!(await Admin.delGroupMember(uid, gid)&&Admin.addGroupMember(uid,gid))) { throw "Erreur dans l'éjection/réadmission du futur admin."; }
        try {
            // Vérifie que l'utilisateur est bien admin (comme dans delGroupMember)
            let la = await Open.getAdmins(gid);
            if (la.includes(uid)) {
                // Supprime tous les administrateurs
                if (!await LDAP.change(ldapConfig.key_id+gid+ldapConfig.dn_groups, "del", ldapConfig.group.admin)) { throw "Erreur dans la suppression de tous les admins pour en supprimer un."; }
                // Les rajoute un par un, sauf pour le supprimé
                la.forEach(id => {
                    if (id!=uid) { Admin.addGroupAdmin(id, gid).then(res => {
                        if (!res) { throw "Erreur dans le réajout d'un des autres admins."; }
                    }); }
                });
            }
            return true;
        }
        catch(err) {
            throw "Erreur lors de l'obtention de la liste des administrateurs d'un groupe.";
        }
    }
    
    //------------------------------------------------------------------------------------------------------------------------
    // Fonction d'édition TBT
    //------------------------------------------------------------------------------------------------------------------------

    /**
     * @summary Fonction qui édite un groupe existant dans le LDAP. Très similaire à {@link User.addGroup}
     * @desc Appelle {@link LDAP.add} bien sûr, mais aussi {@link Admin.addGroupMember} et {@link Admin.addGroupAdmin} en godmode pour gérer les groupes du nouvel utilisateur.
     * @arg {string} gid - Identifiant du groupe à modifier
     * @arg {groupData} data - Dictionnaire des informations utilisateurs au même format que pour {@link User.addGroup} avec tous les champs optionnels.
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async editGroup(gid: string, data: groupData) {
        try {
            // Récupération des anciennes données
            let profil = await Open.peekGroup(gid);
            // Reecriture de profil avec les bons champs
            Object.keys(profil).forEach(keyLDAP => {
                Object.keys(ldapConfig.group).forEach(keyAlias => {
                    ldapConfig.group[keyAlias]=keyLDAP;
                    profil[keyAlias]=profil[keyLDAP];
                });
            });
            // Surcharge des champs à modifier selon data
            Object.keys(data).forEach(key => {
                profil[key]=data[key];
            });
            // Modification propre
            if (!await Admin.delGroup(gid)&&await User.addGroup(profil)) { throw "Erreur de la destruction/recréation du groupe pour le modifier."; }
            return true;
        }
        catch(err) {
            throw "Erreur lors de l'obtention du profil d'un groupe pour le modifier.";
        }
    }
    
    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de suppression TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui supprime un groupe du LDAP.
     * @desc Cette fonction commence par gérer les groupes du membre puis le supprime entièrement.
     * Appelle {@link LDAP.clear} bien sûr, mais aussi {@link Admin.delGroupMember} et {@link Admin.delGroupAdmin} pour gérer les groupes de l'utilisateur sortant.
     * @arg {string} gid - gid de la victime
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async delGroup(gid) {
        try {
            // Gestion des membres et administrateurs d'abord
            let profil = await Open.peekGroup(gid);
            // Ordre important
            profil[ldapConfig.group['admin']].forEach( id => {
                this.delGroupAdmin( id, gid).then(res => { if (!res) { throw "Erreur lors de la suppression d'un admin d'un groupe en cours de suppression."; } });
            });
            profil[ldapConfig.group['member']].forEach(id => {
                this.delGroupMember(id, gid).then(res => { if (!res) { throw "Erreur lors de la suppression d'un membre."; } });
            });
            // Elimination
            if (!await LDAP.clear(ldapConfig.key_id+"="+gid+","+ldapConfig.dn_groups)) { throw "Erreur lors de la suppression de la feuille dans l'arbre des groupes."; }
            return true;
        }
        catch(err) {
            throw "Erreur lors de l'obtention du profil d'un groupe pour le supprimer.";
        }
    }
}

export class Supervisor extends Admin {
    /**
     * @class Cette classe est la classe du super administrateur qui créé et supprime des membres.
     * @summary Constructeur vide.
     * @author hawkspar
     */
    constructor(user) { super(); }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de création TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui créé un nouvel utilisateur dans le LDAP.
     * @desc Appelle {@link LDAP.add} bien sûr, mais aussi {@link User.addGroupMember} et {@link Admin.addGroupAdmin} pour gérer les groupes du nouvel utilisateur.
     * @arg {userFullData} data - Dictionnaire des informations utilisateurs. Des erreurs peuvent apparaître si tous les champs ne sont pas remplis.
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async addUser(data: userFullData) {
        // Calcul d'un dictionnaire d'ajout
        let vals = {};

        // uid de base généré à partir de nom et prénom, plus potentiellement promo et un offset
        // MEF mélange de Promise et de fonction standard
        try {
            Tests.generateUid(data['givenName'],data['lastName'],data['promotion']).then(id => { vals[ldapConfig.key_id]=id; } );
        }
        catch(err) {
            throw "Erreur lors de la génération d'un hruid pour un nouvel utilisateur.";
        }

        // Ecriture de toutes les valeurs directement inscrites dans le LDAP (in pour input)
        // Génère une erreur si un champ n'est pas rempli
        ldapConfig.user.direct_input.forEach(key_att => vals[ldapConfig.user[key_att]]=data[key_att]);

        // Appel à la fonction de base
        if (!await LDAP.add(ldapConfig.key_id+"="+vals[ldapConfig.key_id]+","+ldapConfig.dn_users, vals)) { throw "Erreur de l'ajout de la feuille à l'arbre utilisateur."; }
        
        // Modifications multiples pour avoir plusieurs champs de même type ; boucle sur les attributs multiples
        ldapConfig.user.multiple_input.forEach(key_att => {
            // On rajoute chaque valeur en entrée
            data[key_att].forEach(val => {
                let vals2 = {};
                vals2[ldapConfig.user[key_att]]=val;
                LDAP.change(ldapConfig.key_id+"="+vals[ldapConfig.key_id]+","+ldapConfig.dn_users, "add", vals2).then(res => {
                    if (!res) { throw "Erreur lors de l'ajout d'une valeur pour un champ à valeurs multiples à la feuille du nouvel utilisateur."; }
                });
            });
        });

        // Certains champs nécessitent de petits calculs
        let vals3={};

        // Création d'un nom complet lisible
        vals3[ldapConfig.user['fullName']]=data['givenName']+' '+data['lastName'].toUpperCase();

        // ldapConfiguration du mot de passe utilisateur
        // Le préfixe {CRYPT} signifie que le mdp est hashé dans OpenLDAP voir : https://www.openldap.org/doc/admin24/security.html 
        vals3[ldapConfig.user['password']] = "{CRYPT}"+data['password'];
        
        // Ecriture d'un surnom s'il y a lieu
        if ((data['nickname']!=undefined) && (data['nickname']!='')) {
            vals3[ldapConfig.user['nickname']]=data['nickname'];
        }
        try {
            // Génération id aléatoire unique
            vals3[ldapConfig.user['id']]= await Tests.generateId(ldapConfig.user['id'], ldapConfig.dn_users);
        }
        catch(err) {
            throw "Erreur lors de la génération d'un id numérique pour un nouvel utilisateur.";
        }
        
        // Stockage machine ; dépend du prénom
        vals3[ldapConfig.user['directory']] = '/hosting/users/' + data['givenName'][0];

        // Code root
        vals3[ldapConfig.user['cleanFullName']]=data['fullName'].replace(':', ';').toLowerCase().normalize('UFD');
        
        // Adressage root
        if (data['groups'].includes("on_platal")) { vals3[ldapConfig.user['login']] = "/bin/bash"; }
        else  { vals3[ldapConfig.user['login']] = "/sbin/nologin"; }
        
        // Permissions BR
        vals3[ldapConfig.user['readPerm']] = 'br.*,public.*';
        if (data['readPerm'].length>0) { vals3[ldapConfig.user['readPerm']] += ',' + data['readPerm']; }
        vals3[ldapConfig.user['writePerm']] = 'br.*,!br.blague-du-jour,public.*,!br.campagnekes';
        if (data['writePerm'].length>0) { vals3[ldapConfig.user['readPerm']] += ',' + data['writePerm']; }

        // Valeur nécessaire ASKIP mais inutile
        vals3[ldapConfig.user['idNum']] ='5000';

        // Inscription des valeurs calculées
        if (!await LDAP.change(ldapConfig.key_id+"="+vals[ldapConfig.user['hruid']]+","+ldapConfig.dn_users, "add", vals3)) {
            throw "Erreur lors de l'ajout des valeurs calculées à la feuille du nouvel utilisateur.";
        }

        ["posixAccount", "shadowAccount", "inetOrgPerson", "brAccount"].forEach(cst => {
            let val3={};
            vals3[ldapConfig.user['class']]=cst;
            LDAP.change(ldapConfig.key_id+"="+vals[ldapConfig.user['hruid']]+","+ldapConfig.dn_users, "add", vals3).then(res => {
                if (!res) { throw "Erreur lors de l'ajout d'une valeur constante à la feuille du nouvel utilisateur."; }
            });
        });

        // Utilisation des fonctions adaptées pour assurer la cohérence de l'ensemble
        data['groupsIsMember'].forEach(gid => {
            Admin.addGroupMember(vals[ldapConfig.key_id], gid).then(res => {
                if (!res) { throw "Erreur lors de l'ajout du nouvel utilisateur à un groupe."; }
            });
        });
        data['groupsIsAdmin'].forEach(gid => { 
            Admin.addGroupAdmin( vals[ldapConfig.key_id], gid).then(res => {
                if (!res) { throw "Erreur lors de l'ajout du nouvel utilisateur à un groupe en tant qu'admin."; }
            });
        });

        return true;
    }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de suppression TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui supprime un utilisateur du LDAP.
     * @desc Cette fonction commence par gérer les groupes du membre puis le supprime entièrement.
     * Appelle {@link LDAP.clear} bien sûr, mais aussi {@link Admin.delGroupMember} et {@link Admin.delGroupAdmin} pour gérer les groupes de l'utilisateur sortant.
     * @arg {string} uid - uid de la victime
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async delUser(uid: string) {
        try {
            // Gestion des groupes d'abord
            let profil = await Open.peekUser(uid);
            profil[ldapConfig.user['groups']].forEach(gid => {
                if (Open.isGroupAdmin(uid,gid)) {
                    if (!Admin.delGroupAdmin(uid, gid)) { throw "Erreur lors de la suppression des droits d'admin de l'utilisateur."; }
                }
                if (!Admin.delGroupMember(uid, gid)) { throw "Erreur lors de la suppression de l'appartenace à un groupe de l'utilisateur."; }
            });
        }
        catch(err) {
            throw "Erreur lors de l'obtention des informations de l'utilisateur à supprimer.";
        }
        // Elimination
        if (!LDAP.clear(ldapConfig.key_id+"="+uid+","+ldapConfig.dn_users)) { throw "Erreur lors de la suppression de l'utilisateur."; }
    }
}